@@ -5,7 +5,7 @@ ESM_LIST = [
     "esm1_t12_85M_UR50S",
     "esm1_t6_43M_UR50S",
     "esm1b_t33_650M_UR50S",
-    # "esm_msa1_t12_100M_UR50S",
+    "esm_msa1_t12_100M_UR50S",
 ]
 
 ROSTLAB_LIST = ["Rostlab/prot_bert", "Rostlab/prot_bert_bfd"]
